"""Files helping functions.
"""
import os


def get_file_last_line(filename: str, max_length: int = 80) -> str:
    """Returns the last line from the file, could be used for big files.

    :param filename:
    :param max_length: limits possible size of last line, by default it is \n
    :return:
    """
    with open(filename, "rb") as f_obj:
        f_obj.seek(-max_length, os.SEEK_END)
        return f_obj.readlines()[-1].decode()



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import logging
from decimal import Decimal

from .datamodels import Asset


class AssetBaseClass:
    """Basic implementation of Bot class with Assets.
    """

    def __init__(self, base_asset_name: str, quote_asset_name: str,
                 base_asset_min: Decimal, quote_asset_min: Decimal) -> None:
        """

        :param base_asset_name:
        :param quote_asset_name:
        :param base_asset_min:
        :param quote_asset_min:
        """
        self.symbol = base_asset_name + quote_asset_name
        self.base_asset_name = base_asset_name
        self.quote_asset_name = quote_asset_name

        # Allocating assets:
        setattr(self, self.base_asset_name,  # e.g. BTC
                Asset(name=self.base_asset_name,
                      free=Decimal('0'),
                      locked=Decimal('0'),
                      minimum=base_asset_min))  # e.g. BTC = 0.001
        setattr(self, self.quote_asset_name,  # e.g. BUSD
                Asset(name=self.quote_asset_name,
                      free=Decimal('0'),
                      locked=Decimal('0'),
                      minimum=quote_asset_min))  # e.g. BUSD = $10

        logging.debug('Base asset %s', self.base_asset)
        logging.debug('Quote asset %s', self.quote_asset)

    @property
    def base_asset(self) -> Asset:
        return getattr(self, self.base_asset_name)

    @property
    def quote_asset(self) -> Asset:
        return getattr(self, self.quote_asset_name)

    def sync_balances(self) -> None:
        raise NotImplementedError()



"""Dataclasses helper stands for auto converting field types, whenever it is
possible. Usually Assets represented in JSON and strings formats."""

import logging
from dataclasses import dataclass, fields


@dataclass
class AutoConvertingField:
    """Field auto converting base class."""

    def __post_init__(self):
        for field in fields(self):
            value = getattr(self, field.name)
            if not isinstance(value, field.type):
                logging.debug('Expected %s to be %s, got %s. Converting to %s.'
                              '', field.name, field.type, repr(value),
                              field.type(value))
                setattr(self, field.name, field.type(value))



# coding: utf-8

# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import os
from pymhelper.files import get_file_last_line


def test_get_file_last_line():
    dir_path = os.path.dirname(os.path.realpath(__file__))
    filename = os.path.join(dir_path, 'data', 'test.txt')
    assert get_file_last_line(filename) == ('alias consequatur aut perferendis'
                                            ' doloribus asperiores repellat…')



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

from decimal import Decimal

from pymhelper.numbers import (percentage, round_down, split_amount,
                               prep_range)


def test_percentage():
    assert percentage(percent=1, whole=Decimal('100')) == Decimal('1')
    assert percentage(percent=2, whole=Decimal('100')) == Decimal('2')

    assert percentage(percent=Decimal('1.1111'),
                      whole=Decimal('100')) == Decimal('1.1111')

    assert percentage(percent=3,
                      whole=Decimal('44.64548732')) == Decimal('1.3393646196')


def test_round_down():
    assert round_down(Decimal('1234567890'),
                      6) == Decimal('1234567890').normalize()
    assert round_down(Decimal('123456.7890'),
                      1) == Decimal('123456.7').normalize()
    assert round_down(Decimal('1.234567890'),
                      6) == Decimal('1.234567').normalize()


def test_together():
    assert round_down(
        percentage(percent=3,
                   whole=Decimal('44.64548732')), 2) == Decimal('1.33')


def test_split_amount():
    assert split_amount(min_amount=Decimal('10'),
                        whole=Decimal('44')) == [Decimal('11'), Decimal('11'),
                                                 Decimal('11'), Decimal('11')]
    assert split_amount(min_amount=Decimal('0.002'),
                        whole=Decimal('0.003')) == [Decimal('0.003')]
    assert split_amount(min_amount=Decimal('0.001'),
                        whole=Decimal('0.002')) == [Decimal('0.001'),
                                                    Decimal('0.001')]


def test_prep_range():
    assert prep_range(min_value=Decimal('1'),
                      max_value=Decimal('2'),
                      length=4) == [Decimal('1'), Decimal('1.25'),
                                    Decimal('1.5'), Decimal('1.75')]



# pylint: disable=missing-module-docstring
from .base import AssetBaseClass
from .datamodels import Asset
from .exceptions import AmountTooSmall

__all__ = ('Asset', 'AssetBaseClass', 'AmountTooSmall')



# pylint: disable=missing-module-docstring
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

from dataclasses import dataclass
from decimal import Decimal

from pymhelper.dataclasseshelper import AutoConvertingField


@dataclass
class TestDataClass(AutoConvertingField):
    dec: Decimal
    num: int
    string: str


def test_auto_convert():
    tdc = TestDataClass(dec='5.5001', num='5', string=5)
    assert isinstance(tdc.dec, Decimal)
    assert tdc.dec == Decimal('5.5001')
    assert isinstance(tdc.num, int)
    assert tdc.num == 5
    assert isinstance(tdc.string, str)
    assert tdc.string == '5'



"""Set of function related to numbers, math, percentages.
"""
from decimal import Decimal, ROUND_DOWN
from typing import Union, List


def percentage(*, percent: Union[int, Decimal], whole: Decimal) -> Decimal:
    """Returns percentage of a number.

    :param percent: for example Decimal('105.95'), means 105.95%.
    :param whole:
    :return:
    """
    return Decimal((percent * whole) / 100)


def round_down(number: Decimal, decimals: int = 6) -> Decimal:
    """Round down decimal.

    :param number:
    :param decimals:
    :return:
    """
    return number.quantize(Decimal(10) ** -decimals,
                           rounding=ROUND_DOWN).normalize()


def split_amount(*, min_amount: Decimal, whole: Decimal) -> List[Decimal]:
    """Splits number into almost equal parts of the whole number which are
    greater then min_amount.

    For example 56 / 10 = 11.2.

    :param min_amount:
    :param whole:
    :return:
    """
    parts = whole // min_amount
    return [whole / parts for _ in range(int(parts))]


def prep_range(*,
               min_value: Decimal,
               max_value: Decimal,
               length: int) -> List[Decimal]:
    """Example:

    prep_range(min_value=Decimal('0.001'), max_value=Decimal('0.005'),
               length=5)
        [
            Decimal('0.0010'), Decimal('0.0018'), Decimal('0.0026'),
            Decimal('0.0034'), Decimal('0.0042')
        ]

    :param min_value:
    :param max_value:
    :param length:
    :return:
    """
    step = (max_value - min_value) / length
    return [min_value + Decimal(step) * i for i in range(length)]



"""Assets represented in dataclasses format.
"""

from dataclasses import dataclass
from decimal import Decimal

from pymhelper.dataclasseshelper import AutoConvertingField
from pymhelper.numbers import round_down


@dataclass
class Asset(AutoConvertingField):
    """Base or quote asset.
    """
    name: str
    free: Decimal
    locked: Decimal

    minimum: Decimal
    step: Decimal = Decimal('0')

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return (f'{self.name} free / locked: {self.free} / '
                f'{self.locked}')

    @property
    def is_enough(self) -> bool:
        """Check is free asset enough to buy or sell.
        :return:
        """
        return self.free > self.minimum + self.step

    def round(self, precision: int = 6) -> Decimal:
        """Returns rounded free asset.

        :param precision:
        :return:
        """
        return round_down(self.free, precision)



# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import logging
from decimal import Decimal

from pymhelper.assets import Asset

logging.basicConfig(level=logging.DEBUG)


def test_asset_creation():
    btc = Asset(name='BTC', free=Decimal('1234.5678'), locked=Decimal('0'),
                minimum=Decimal('0'))
    assert btc.free == Decimal('1234.5678')
    assert btc.name == 'BTC'
    assert str(btc) == 'BTC free / locked: 1234.5678 / 0'


def test_asset_creation_from_dict():
    btc = Asset(**dict(name='BTC', free='1234.5678', locked='0', minimum='10'))
    assert btc.free == Decimal('1234.5678')
    assert btc.name == 'BTC'
    assert str(btc) == 'BTC free / locked: 1234.5678 / 0'
    assert btc.minimum == Decimal('10')


def test_auto_field_convert():
    btc = Asset(name='BTC', free=22, locked='0.001', minimum='10')
    assert btc.free == Decimal('22')
    assert btc.locked == Decimal('0.001')
    assert btc.minimum == Decimal('10')


def test_free_round():
    busd = Asset(name='BUSD', free=Decimal('12.3456789'), locked=Decimal('0'),
                 minimum=Decimal('10'))
    assert busd.round() == Decimal('12.345678')
    assert busd.round(2) == Decimal('12.34')


def test_is_enough():
    btc = Asset(name='BTC', free=10, locked=0, minimum=10, step=1)
    assert btc.is_enough is False
    btc = Asset(name='BTC', free=11, locked=0, minimum=10, step=1)
    assert btc.is_enough is False
    btc = Asset(name='BTC', free=11.1, locked=0, minimum=10, step=1)
    assert btc.is_enough is True
